/* Autor: Ylva Brandt, Juni 2020
* Dieses Werk ist lizenziert unter einer Creative Commons Namensnennung - Nicht-kommerziell - Weitergabe unter gleichen Bedingungen 4.0 International Lizenz (http://creativecommons.org/licenses/by-nc-sa/4.0/).
* Sie erlaubt Download und Weiterverteilung des vollständigen Werkes unter Nennung meines Namens, jedoch keinerlei kommerzielle Nutzung. Die Weitergabe von Bearbeitungen dieses Werkes ist erlaubt, 
* solange ich als Urheber des Originals genannt werde und die auf diesem Werk basierenden neuen Werke unter denselben Bedingungen veröffentlicht werden.
*/ 

/* Das Programm verwendet die Datenstruktur ArrayList als dynamische Reihung. Eine Lösung, welche die im Abitur vorgesehenen Datenstrukturen 
 Schlange bzw. Dynamische Reihung verwendet, wird ergänzt, sobald eine passende Bilbiothek bereitsteht. 
 Eine entsprechende Bibliothek ist seitens des Landesnetzwerkes Informatik in Planung.
 */


/* Anleitung: 
* Mit der linken Maustaste kann ein Auswahlbereich aufgezogen werden. Andernfalls wird die Funktion auf das gesamte Bild angewendet.
* Mit der rechten Maustaste kann ein Pixel als Referenzpunkt für die Zauberstabfunktion ausgewählt werden.
* Auswahl der Funktion über Tasten:
* Taste r: Zurücksetzen des Bildes und des Auswahlbereichs
* Taste z: Alle Bildpunkte mit einem ähnlichen RGB-Wert zum Referenzpunkt werden durch Weiß ersetzt. Der Toleranzbereich kann in Prozent ausgewählt werden. 
* Taste i: Ausgehend vom Referenzpunkt werden alle angrenzenden Bildpunkte mit einem ähnlichen RGB-Wert durch Weiß ersetzt.
* Taste s: Das aktuell angezeigte Bild wird gespeichert.
*/

import static javax.swing.JOptionPane.*;

PImage meinBild; //Variable, die das zu bearbeitende Bild aufnimmt
int zaehler=1; // Zaehler, um mehrere Versionen des Bildes abspeichern zu können
color refColor; //Referenz-RGB-Wert für die Zauberstabfunktion
int x1 = 0; //Koordinaten der linken oberen Ecke des Auswahlbereichs
int y1 = 0;
int x2; //Koordinaten der rechten unteren Ecke des Auswahlbereichs
int y2;
int startX, startY; //Koordinanten des ausgewählten Referenzpunktes. Bei dem iterativen Löschen von zusammenhägnenden Pixeln ähnlicher Farbe sind dies die Startkoordinaten


void setup(){
  size(800, 450); //Die Größe des Fensters wird auf die Größe des Bildes festgelegt.
  x2 = width; //Initialwerte des Auswahlbereichs für das gesamte Fenster, x2 und y2 enthalten die rechte, untere Ecke des gesamten Bildes
  y2 = height;
  meinBild = loadImage("koralle_800x450.jpg"); //Das Bild wird in der globalen Variablen meinBild gespeichert
  image(meinBild, 0, 0); //Das Bild in der Variablen meinBild wird im Programmfenster angezeigt.
  loadPixels(); //Die RGB-Werte des aktuell angezeigten Bildes werden in die Systemvariable pixels (Reihung vom Typ color) geladen
}

void draw()
{
  if(mousePressed && mouseButton==LEFT){//Anzeigen des vorläufig ausgewählten Rechtecks
    noFill(); //es wird nur die Umrandung gezeichnet
    updatePixels(); //das vorherige Auswahlrechteck wird durch das aktuelle Bild übermalt
    //vom Startpunkt, an dem die Maus gedrückt wurde, wird zur aktuellen Mausposition ein Rechteck aufgezogen.
    //Die Koordinaten der linken, oberen Ecke werden dabei als Minimum der x- bzw. y-Koordinaten bestimmt.
    rect(min(x1,mouseX), min(y1,mouseY), abs(mouseX-x1), abs(mouseY-y1));  
  }
}

void mouseClicked(){
 if(mouseButton == RIGHT){ //Auswahl des Referenzpunktes mit der rechten Maustaste
  refColor = pixels[mouseX + mouseY*width]; //Der Index des Referenzpunktes berechnet sich aus x + y * Breite des Bildes
  println(refColor);
  startX = mouseX; //Die Koordinaten des Refernzpunktes werden für die Auswahl der zusammenhängenden Pixel gespeichert 
  startY = mouseY;
 }
}

//mit der linken Maustaste kann ein rechteckiger Auswahlbereich festgelegt werden.
void mousePressed(){
 if(mouseButton == LEFT){
  //In x1 und y1 werden die Koordinaten des Startpunktes für den Auswahlbereich gespeichert. 
  x1 = mouseX;
  y1 = mouseY;
 }
}

void mouseReleased(){//Auswahl des Rechtecks abgeschlossen
  if(mouseButton == LEFT){
  //Die Koordinaten werden so verteilt, dass in x1, y1 die Koordinaten der linken, oberen Ecke gespeichert werden
  //und in x2, y2 die Koordinaten der rechten, unteren Ecke  
  //Falls die Koordinaten außerhalb des Fensters liegen, werden sie entsprechend angepasst
  int xalt = x1;
  int yalt = y1;
  x1 = min(xalt, mouseX);
  if(x1<0) x1 = 0;
  y1 = min(yalt, mouseY);
  if(y1<0) y1 = 0;
  x2 = max(xalt,mouseX);
  if(x2>width) x2 = width;
  y2 = max(yalt,mouseY);
  if(y2>height) y2 = height;
  System.out.println("Released: 1:( " + x1 + ", " + y1 + ") 2: ("+ x2 + ", " + y2 + ")"); //Ausgabe zur Kontrolle
  }
}


void keyPressed(){// über die Tasten wird die Funktion ausgewählt, die auf das Bild angewendet werden soll
  switch(key){
    case 'r': reset();
              break;
    case 'z': zauberstab();   //Anwenden der Zauberstab-Funktion im ausgewählten Bereich
              break;
    case 'i': zauberstabIterativ(); //Anwenden der Zauberstab-Funktion auf zusammenhängende Pixel ausgehend vom Referenzpixel
              break;
    case 's': save("bildBearbeitung" + zaehler+".png"); //abspeichern der aktuellen Version des Bildes
              zaehler++;
              break;
      
  }
}

   void reset(){//Die Anzeige auf die ursprüngliche Version des Bildes zurücksetzen
     image(meinBild, 0, 0);
     loadPixels(); //die aktuellen RGB-Werte werden in eine Reihung vom Typ color geladen. Diese Reihung steht als Systemvariable pixels zur Verfügung
     //der Auswahlbereich wird auf die gesamte Fenstergröße gesetzt
     x1 = 0;
     y1 = 0;
     x2 = width;
     y2 = height;
   }
    
 //Alle Pixel die einen zum Referenzwert ähnlichen RGB-Wert haben, werden weiß gefärbt
 void zauberstab(){
   try{
    float toleranz = -1;
    while(toleranz < 0 || toleranz > 100){
      toleranz = Float.parseFloat(showInputDialog("Bitte gib den Toleranzwert in Prozent als Zahl zwischen 0 und 100 ein!")); //Abfrage des Toleranzwertes in Prozent
    }
    int tWert = round((255 * toleranz/100)); //Umrechnung des Prozentwertes in einen absoluten Wert
    for(int y = y1; y < y2; y++){ //Iteration durch alle Pixel des Auswahlbereichs
     for(int x = x1; x < x2; x++){
       color col = pixels[x+y*width]; //Auslesen des RGB-Wertes des aktuellen Bildpunktes
       if(vergleicheRGBWerte(refColor, col, tWert)){ //Weiß-Färben, wenn der Bildpunkt das Ähnlichkeitskriterium erfüllt
          pixels[x+y*width] = color(255,255,255);
       }
     }
    }
    updatePixels(); // Die aktuellen Werte in der Reihung pixels werden für die Anzeige des Bildes übernommen.  
   }catch(NumberFormatException e){
     println("Bitte gib eine Zahl ein.");
   }
 }
   
 /* Ausgehend vom ausgewählten Referenzpixel werden alle ähnlichen Pixel in der Umgebung weiß gefärbt. 
  * Dazu werden alle acht Nachbarpixel des Referenzpixels betrachtet. Wenn ein Pixel einen ähnlichen RGB-Wert hat, 
  * wird es weiß gefärbt und seine Koordinaten an eine dynamische Reihung (ArrayList) angehängt. Für alle Pixel in der dynamsischen Reihung werden die Nachbarpixel auf die gleiche
  * Weise behandelt.
  * Da eine rekursive Lösung beim Testen die Grenzen der Rekursionstiefe von Processing überschritten hat, wurde hier eine iterative Lösung gewählt 
 */  
 void zauberstabIterativ(){
  try{
   float toleranz = -1;
   while(toleranz < 0 || toleranz > 100){
      toleranz = Float.parseFloat(showInputDialog("Bitte gib den Toleranzwert in Prozent als Zahl zwischen 0 und 100 ein!")); //Abfrage des Toleranzwertes in Prozent
   }
  int tWert = round(255 * toleranz/100); //Umrechnung des Prozentwertes in einen absoluten Wert
  ArrayList<int[]> koordinatenListe = new ArrayList<int[]>();
  pixels[startX+startY*width] = color(255, 255, 255);  //Startpixel wird weiß gefärbt
  int[] temp = {startX, startY};  //die Koordinaten eines Pixels werden als Reihung mit zwei Einträgen vom Typ Ganzzahl gespeichert
  koordinatenListe.add(temp); //anhängenden der Koordinaten des Startpixels an die Liste
  while(!koordinatenListe.isEmpty()){//so lange noch Koordinaten in der dynamischen Reihung vorhanden sind, werden diese entnommen und die Umgebungspixel untersucht
     int[] koordinaten = koordinatenListe.get(0); //Da der Rückgabetyp von getItem() Object ist, wird hier mittels (int[]) ein Casting vorgenommen
     int x = koordinaten[0];
     int y = koordinaten[1];
     koordinatenListe.remove(0);
     for(int i = x-1; i <= x+1; i++){ //untersuchen aller Umgebungspixel 
       for(int j = y-1; j <= y+1; j++){   
         if(i+j*width >=0 && i+j*width < pixels.length && !(i==x && j == y)){//testen, ob die Koordinaten zu einem Nachbarpixel gehören und innerhalb des Bildes liegen
            color col = pixels[i+j*width];
            if(vergleicheRGBWerte(refColor, col, tWert)){ //Nur farbliche übereinstimmende Pixel werden weiß gefärbt und ihre Koordinaten in die dynamische Reihung eingefügt
               pixels[i+j*width] = color(255, 255, 255);
               int[] temp2 = {i,j};
               koordinatenListe.add(temp2);
           }
         }
       }
     }
   }
   updatePixels();// Die aktuellen Werte in der Reihung pixels werden für die Anzeige des Bildes übernommen. 
  }
   catch(NumberFormatException e){
     println("Bitte gib eine Zahl ein.");
   }
 }
 
  
 ///Hilfmethode zum Überprüfen ob zwei RGB-Werte unter Berücksichtigung des Toleranzwertes farblich übereinstimmen sind 
 boolean vergleicheRGBWerte(color col1, color col2, int abweichung){
    boolean gleich = false;
    //Die Rot-, Grün- und Blauwerte werden jeweils verglichen. Dabei muss die Differenz der Werte kleiner als der Toleranzwert sein, damit die Werte als übereinstimmend gewertet werden.   
    if((abs(red(col1) - red(col2)) < abweichung) &&
        (abs(green(col1) - green(col2)) < abweichung) && 
        (abs(blue(col1) - blue(col2)) < abweichung)){
        gleich = true;
    }
    return gleich;
}
