/*Autor: Ylva Brandt, Mai 2020
* Dieses Werk ist lizenziert unter einer Creative Commons Namensnennung - Nicht-kommerziell - Weitergabe unter gleichen Bedingungen 4.0 International Lizenz (http://creativecommons.org/licenses/by-nc-sa/4.0/).
* Sie erlaubt Download und Weiterverteilung des vollständigen Werkes unter Nennung meines Namens, jedoch keinerlei kommerzielle Nutzung. Die Weitergabe von Bearbeitungen dieses Werkes ist erlaubt, solange ich als Urheber des Originals genannt werde und die auf diesem Werk basierenden neuen Werke unter denselben Bedingungen veröffentlicht werden.

* Ausgenommen von der Lizenz ist das im Ordner data beiliegende Portraitfoto (portrait_600x600.jpg). Dieses unterliegt dem Urheberrecht der Autorin.
*/

/*Anleitung:
r: Zurücksetzen des Bildes
m: Erstellen der Vorlage für das Mosaik ohne Umrandung der Bausteine
u: Erstellen der Vorlage für das Mosaik mit Umrandung der Bausteine
b: Erstellen des Mosaiks aus Bausteinfotos 
s: Abspeichern der aktuellen Version des Bildes
*/

import static javax.swing.JOptionPane.*;

PImage meinBild; //Variable, die das zu bearbeitende Bild aufnimmt
PImage[] bausteine = new PImage[5]; //Reihung vom Typ PImage, die je ein Bild des Bausteine für das Baustein-Mosaik enthält  
boolean rand = false; //Hier wird sich gemerkt, ob die Bausteine in der Vorlage von grünen Linien abgegrenzt werden sollen
int zaehler=1; // Zaehler, um mehrere Versionen des Bildes abspeichern zu können


void setup(){
  size(600, 600); //Die Größe des Fensters wird auf die Größe des Bildes festgelegt.
  meinBild = loadImage("portrait_600x600.jpg"); //Das Bild wird geladen und in der globalen Variablen meinBild gespeichert
  image(meinBild, 0, 0); //Das Bild in der Variablen meinBild wird im Programmfenster angezeigt.
  loadPixels(); //die aktuellen RGB-Werte werden in eine Reihung vom Typ color geladen. Diese Reihung steht als Variable pixels zur Verfügung
  for(int i = 0; i < bausteine.length; i++){
   bausteine[i] = loadImage("baustein" + i + ".png"); //Die Bilder der Bausteine werden in die Bilder-Reihung geladen
  }
}

void draw()// Diese Methode muss enthalten sein, damit die Anzeige des Bildes akutalisiert wird. 
{
}

void keyPressed(){// über die Tasten wird die Funktion ausgewählt, die auf das Bild angewendet werden soll
  switch(key){
    case 'r': reset();  //Zurücksetzen des Bildes
              break;
    case 'm': rand = false;
              mosaik(); //Erstellen der Vorlage für das Mosaik ohne Umrandung der Bausteine
              break;
    case 'u': rand = true; //Erstellen der Vorlage für das Mosaik mit Umrandung der Bausteine
              mosaik();
              break;
    case 'b': mosaikBausteine(); //Erstellen des Mosaiks aus Bausteinfotos 
              break;
    case 's': save("bildBearbeitung" + zaehler+".png"); //abspeichern der aktuellen Version des Bildes
              zaehler++;
              break;  
  }
}

//Erzeugen einer Vorlage für das Mosaiks
 void mosaik(){
   //Definieren der Farben, welche die Bausteine haben können
   color weiss = color(255, 255, 255);
   color hellgrau = color(180, 180, 180);
   color dunkelgrau = color(100, 100, 100);
   color schwarz = color(0, 0, 0);
   color hintergrund = color(255, 255, 0);
   //Es wird von einem einheitlich gefärbten Hintergrund ausgegangen. Der entsprechende Farbwert wird dem ersten Pixel der Reihung pixels entnommen und als Referenzwert gespeichert. 
   color hintergrundOrg = pixels[0];
   //Der Anwendet wählt die Breite der Bausteine in Pixeln aus.
   int pixelbreite = Integer.parseInt(showInputDialog("Bitte gib die Breite des Bausteins in Pixeln an."));
   //Um mit der Pixelbreite besser arbeiten zu können, wird die Angabe ggf. auf die nächste gerade Zahl verkleinert 
   if((pixelbreite % 2) == 0) pixelbreite--;
   int halbePixelbreite = pixelbreite/2;
   //Für jeden Baustein wird der Farbwert des Pixels, das sich in der Mitte des Bausteins befindet, betrachtet. Daher starten die Indizes bei der halben Pixelbreite
   //Zwei geschachtelte Schleifen iterieren durch die x- und die y- Koordinate der Pixel in der Mitte eines Bausteins
   for(int i = halbePixelbreite; i < width; i = i+pixelbreite){
     for(int j = halbePixelbreite; j < height; j = j+pixelbreite){
        color col = pixels[i+j*width]; //Auslesen des Farbwerts
        //Wenn der Farbwert dem Farbwert des Hintergrunds entspricht, werden alle Pixel, die zum Baustein gehören in der Farbe des Hintergrundbausteins (hier: gelb) gefärbt
        if(col == hintergrundOrg){
           for(int x = i-halbePixelbreite; x <= i+halbePixelbreite; x++){
             for(int y = j-halbePixelbreite; y <= j+halbePixelbreite; y++){
               if(x >= 0 && y >= 0 && x < width && y < height){//Testen, ob sich die Koordinaten noch innerhalb der Reichung des Bildes befinden.
                 pixels[x+y*width] = hintergrund;
               }
             }
           }
        }else{
          //Für alle Pixel, die nicht zum Hintergrund gehören, wird der Durchschnitt der Farbanteile als Helligkeit bestimmt
          float helligkeit = (red(col) + green(col) + blue(col))/3;
          
         //Alternativer Code, der alle Pixel der Gruppe beim Bestimmen der Helligkeit berücksichtigt.
          /*float helligkeit = 0;
          int zaehler = 0;
          for(int x = i-halbePixelbreite; x <= i+halbePixelbreite; x++){
             for(int y = j-halbePixelbreite; y <= j+halbePixelbreite; y++){
               if(x >= 0 && y >= 0 && x < width && y < height){//Testen, ob sich die Koordinaten noch innerhalb der Reichung des Bildes befinden.
                 zaehler++;
                 color col2 = pixels[x+y*width];
                 helligkeit = helligkeit + ((red(col2) + green(col2) + blue(col2))/3);
               }
             }
           }
           helligkeit = helligkeit/zaehler;
           */
           
          //Der Variablen graustufe wird abhängig von der Helligkeit der passende Grauton für den Baustein zugewiesen.
          color graustufe = weiss;
          if(helligkeit < 50){
            graustufe = schwarz;
          }else{
            if(50 <= helligkeit && helligkeit < 95){
              graustufe = dunkelgrau;
            }else{
              if(95 <= helligkeit && helligkeit < 185){
                graustufe = hellgrau;
              }
            }
          }
          //Alle Pixel, die zum Baustein gehören, werden in der passenden graustufe gefärbt
          for(int x = i-halbePixelbreite; x <= i+halbePixelbreite; x++){
             for(int y = j-halbePixelbreite; y <= j+halbePixelbreite; y++){
               if(x >= 0 && y >= 0 && x < width && y < height){
                 pixels[x+y*width] = graustufe;
               }
             }
           }
        }
      }
   }
   //Wenn die Bausteine zur Abgenzung umrandet werden sollen, werden hier die entsprechenden Linien eingezeichnet.
   //Dazu werden direkt die Pixel in der Reihung gefärbt. Alternativ könnten tatsächlich Linien eingezeichnet werden.
   if(rand){
     for(int i = 0; i < width; i = i+pixelbreite){
       for(int j = 0; j < height; j++){
         pixels[i+j*width] = color(0, 255, 0);
       }
     }
     for(int i = 0; i < width; i++){
       for(int j = 0; j < height; j = j + pixelbreite){
         pixels[i+j*width] = color(0, 255, 0);
       }
     }
   }
   
   updatePixels(); // Die aktuellen Werte in der Reihung pixels werden für die Anzeige des Bildes übernommen.  
 
}
  

 //Erzeugen eines Mosaiks aus Bausteinfotos
 //Der Algorithmus ist ähnlich zu dem Algorithmus für das Erzeugen der Vorlage. 
 //Allerdings wird hier von jedem Baustein das Pixel in der linken, oberen Ecke betrachtet.
 //Die Ergebnisse können daher von einander abweichen
 //Außerdem wird nicht die Reihung pixels verändert, sondern das Bild aus einzelnen Bildern im Programmfenster zusammengesetzt
 void mosaikBausteine(){
   int pixelbreite = Integer.parseInt(showInputDialog("Bitte gib die Breite des Bausteins ein."));
   color hintergrundOrg = pixels[0];
   for(int i = 0; i < width; i = i+pixelbreite){
     for(int j = 0; j < height; j = j+pixelbreite){
        color col = pixels[i+j*width]; 
           if(col == hintergrundOrg){
               image(bausteine[4], i, j, pixelbreite, pixelbreite); //Einzeichnen der Hintergrundbausteins
        }else{
          float helligkeit = (red(col) + green(col) + blue(col))/3;
          //In der Variablen graustufe wird der Index des passenden Bausteins in der Reihung bausteine gespeichert
          color graustufe = 0;
          if(helligkeit < 50){
            graustufe = 3;
          }else{
            if(50 <= helligkeit && helligkeit < 95){
              graustufe = 2;
            }else{
              if(95 <= helligkeit && helligkeit < 185){
                graustufe = 1;
              }
            }
          }
          image(bausteine[graustufe], i, j, pixelbreite, pixelbreite); //Einzeichnen des passenden Bausteins
        }
      }
   }
}
  
//Die Anzeige auf die ursprüngliche Version des Bildes zurücksetzen
void reset(){
   image(meinBild, 0, 0, width, height);   
   loadPixels();
}
